"""Module for log functions"""

import os
import re
import math
import shutil
from datetime import datetime
from colorama import Fore, Style
from version import PLOOSH_VERSION


class Log:
    """Log class contain all functions to log"""

    @staticmethod
    def init():
        Log.LEVELS_PRINT = {
            "INFO": Fore.GREEN,
            "WARN": Fore.YELLOW,
            "ERRO": Fore.RED
        }

        Log.CONSOLE_WIDTH = shutil.get_terminal_size(fallback=(120, 50)).columns
        Log.CONSOLE_WIDTH_GAP = 29
        Log.CONSOLE_LOG_SPACE = Log.CONSOLE_WIDTH - Log.CONSOLE_WIDTH_GAP

        Log.LOGS_FOLDER = "./logs"
        Log.LOGS_PATH = f"{Log.LOGS_FOLDER}/ploosh_{datetime.now().strftime('%Y%m%d%H%M%S')}.log"

        os.makedirs(Log.LOGS_FOLDER, exist_ok=True)

    @staticmethod
    def print(message:str, level:str = "INFO", filler:str = "."):
        """Print a message with all metadata informations"""

        date_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        count_filler = 1 if message.count("[...]") == 0 else message.count("[...]")
        
        raw_message = re.sub(r'[^\w ]*[\d]+m', '', message)
        print_length = len(raw_message)
        feed_characters = filler * math.trunc((Log.CONSOLE_LOG_SPACE - print_length + (5 * count_filler)) / count_filler)
        message = message.replace("[...]", feed_characters)

        rows_to_print = [message]
        # coloration will be disabled for multi ligne message
        if print_length > Log.CONSOLE_LOG_SPACE or "\n" in message:
            rows_to_print = []
            message_rows = raw_message.split("\n")
            for row in message_rows:
                rows_count = math.ceil(len(row) / Log.CONSOLE_LOG_SPACE)
                for i in range(0, rows_count):
                    start = i * Log.CONSOLE_LOG_SPACE
                    end = (i + 1) * Log.CONSOLE_LOG_SPACE
                    rows_to_print.append(row[start:end])

        rows_to_print = [f"{Fore.CYAN}[{date_time}] {Log.LEVELS_PRINT[level]}[{level}]{Style.RESET_ALL} {row}{Style.RESET_ALL}" for row in rows_to_print]

        for row in rows_to_print:
            print(row)

        

        with open(Log.LOGS_PATH, "a", encoding="UTF-8") as f:
            log_text = "\r\n".join(rows_to_print) + "\r\n"

            for key in Fore.__dict__:
                log_text = log_text.replace(Fore.__dict__[key], "")

            for key in Style.__dict__:
                log_text = log_text.replace(Style.__dict__[key], "")

            f.write(log_text)

    @staticmethod
    def print_error(message:str):
        """Print an error message with all metadata informations"""
        Log.print(message, "ERRO")

    @staticmethod
    def print_warning(message:str):
        """Print an warning message with all metadata informations"""
        Log.print(message, "WARN")

    @staticmethod
    def print_logo():
        """Print the ATF logo"""

        Log.print(f"[...]", filler="~")
        Log.print(f"[...]       .__                      .__     [...]", filler=" ")
        Log.print(f"[...]______ |  |   ____   ____  _____|  |__  [...]", filler=" ")
        Log.print(f"[...]\____ \|  |  /  _ \ /  _ \/  ___|  |  \ [...]", filler=" ")
        Log.print(f"[...]|  |_> |  |_(  <_> (  <_> \___ \|   Y  \[...]", filler=" ")
        Log.print(f"[...]|   __/|____/\____/ \____/____  |___|  /[...]", filler=" ")
        Log.print(f"[...]|__|                          \/     \/ [...]", filler=" ")
        Log.print(f"[...]Automatized Testing Framework (v {PLOOSH_VERSION})[...]", filler=" ")
        Log.print(f"[...]", filler=" ")
        Log.print(f"[...]https://github.com/CSharplie/ploosh", filler=" ")
        Log.print(f"[...]", filler="~")

def print_compare_state(current_case):
    state = current_case.state.upper()
    state_matrix = {
        "FAILED" : { "color": Fore.YELLOW, "function": Log.print_warning },
        "ERROR" : { "color": Fore.RED, "function": Log.print_error },
        "PASSED" : { "color": Fore.GREEN, "function": Log.print },
    }
    state_item = state_matrix[state]
    state_item["function"](f"Compare state: {state_item['color']}{state}")
    
    if state != "PASSED":
        state_item["function"](f"Error type   : {state_item['color']}{current_case.error_type.upper()}")
        state_item["function"](f"Error message: {state_item['color']}{current_case.error_message}")

def print_summary(cases, statistics):
    for case_name in cases:
        state = cases[case_name].state
        color = Fore.CYAN

        if state == "error": color = Fore.RED
        if state == "passed": color = Fore.GREEN
        if state == "failed": color = Fore.YELLOW

        if state == "notExecuted": state = "skipped"

        Log.print(f"{case_name} [...] {color}{state.upper()}")

    message = f"passed: {Fore.GREEN}{statistics.passed}{Style.RESET_ALL}, "
    message += f"failed: {Fore.YELLOW}{statistics.failed}{Style.RESET_ALL}, "
    message += f"error: {Fore.RED}{statistics.error}{Style.RESET_ALL}, "
    message += f"skipped: {Fore.CYAN}{statistics.not_executed}{Style.RESET_ALL}"

    Log.print(message)


"""Automatized Testing Framework"""

import sys
from colorama import Fore, Style
from pyspark.sql import SparkSession
from case import StateStatistics
from logs import Log, print_compare_state, print_summary
from connectors import get_connectors
from exporters import get_exporters
from parameters import Parameters
from configuration import Configuration

def load_data(current_case, process_type, statistics):
    """load data from source or expected"""

    try:
        current_case.load_data(process_type)
        return True
    except Exception as e:
        current_case.load_data_error(process_type, str(e))
        current_case.calculate_durations()
        statistics.add_state(current_case.state)
        Log.print_error(str(e))

    return False

def compare_data(current_case, statistics, spark_session):
    """Compare data between source and expected"""

    try:
        if current_case.source.connector.is_spark and current_case.expected.connector.is_spark:
            current_case.compare_dataframes_with_spark(spark_session)
        else:
            current_case.compare_dataframes()
        statistics.add_state(current_case.state)
        return True
    except Exception as e:
        current_case.compare_dataframes_error(str(e))
        current_case.calculate_durations()
        statistics.add_state(current_case.state)
        Log.print_error(str(e))

    return False

def execute(args=None, spark_session=None):
    """Main function"""

    Log.init()
    Log.print_logo()

    statistics = StateStatistics()

    Log.print(f"{Fore.CYAN}Initialization[...]")
    try:
        if args is None:
            parameters = Parameters(sys.argv)
        else:
            parameters = Parameters(args)

        if parameters.spark_mode is True and spark_session is None:
            Log.print("Start spark session")
            spark_session = SparkSession.builder \
                .master("local") \
                .appName("ploosh") \
                .getOrCreate()

        Log.print("Load connectors")
        connectors = get_connectors(spark_session)
        Log.print("Load exporters")
        exporters = get_exporters()

        Log.print("Load configuration")
        configuration = Configuration(parameters, connectors, exporters)
        cases = configuration.get_cases()
    except Exception as e:
        Log.print_error(str(e))
        exit(1)

    Log.print(f"{Fore.CYAN}Start processing tests cases[...]")
    for i, case_name in enumerate(cases):
        current_case = cases[case_name]
        
        if current_case.disabled:
            Log.print(f"{Fore.MAGENTA}{case_name} [...] ({i + 1}/{len(cases)}) - Skipped")
            statistics.add_state(current_case.state)
            continue

        Log.print(f"{Fore.MAGENTA}{case_name} [...] ({i + 1}/{len(cases)}) - Started")

        Log.print("Load source data")
        if not load_data(current_case, "source", statistics):
            continue

        Log.print("Load expected data")
        if not load_data(current_case, "expected", statistics):
            continue

        Log.print("Compare source and expected data")
        if not compare_data(current_case, statistics, spark_session):
            continue

        print_compare_state(current_case)

        current_case.calculate_durations()

    Log.print(f"{Fore.CYAN}Export results[...]")
    configuration.exporter.export(cases)
    Log.print(f"{Fore.CYAN}Summary[...]")

    print_summary(cases, statistics)

    if statistics.error > 0 and parameters.failure_on_error:
        exit(1)


# pylint: disable=R0903
"""Connector to return empty"""

import pandas as pd
from connectors.connector import Connector

class ConnectorEmpty(Connector):
    """Connector to return empty"""
    def __init__(self):
        self.name = "EMPTY"
        self.connection_definition = []
        self.configuration_definition = []

    def get_data(self, configuration: dict, connection: dict):
        """Return empty value"""
        df = pd.DataFrame()
        return df



"""Initialization for command line"""

import os
import sys

sys.path.append(os.path.dirname(__file__))

from execute import execute

def execute_cases(cases = None, connections = None, spark = None, spark_session = None, filter = None, path_output = None):
    args = ["ploosh"]

    if cases is not None:
        args.append("--cases")
        args.append(cases)

    if connections is not None:
        args.append("--connections")
        args.append(connections)

    if spark is not None:
        args.append("--spark")
        args.append(spark)

    if filter is not None:
        args.append("--filter")
        args.append(filter)

    if path_output is not None:
        args.append("--output")
        args.append(path_output)

    execute(args, spark_session)


"""Export test case result to CSV format"""
import csv
import os
from exporters.exporter import Exporter

class ExporterCSV(Exporter):
    """Export test case result to CSV format"""

    def __init__(self):
        self.name = "CSV"

    def export(self, cases:dict):
        """Export test case"""

        output_file = f"{self.output_path}/csv/test_results.csv"

        data = [[
               "name",
               "state",
               "source_start",
               "source_end",
               "source_duration",
               "expected_start",
               "expected_end",
               "expected_duration",
               "compare_start",
               "compare_end",
               "compare_duration",
               "error_type",
               "error_message",
        ]]
        for name in cases:
            case = cases[name]

            case_data = [
                name,
                case.state,
                Exporter.date_to_string(case.source.duration.start),
                Exporter.date_to_string(case.source.duration.end),
                case.source.duration.duration,
                Exporter.date_to_string(case.expected.duration.start),
                Exporter.date_to_string(case.expected.duration.end),
                case.expected.duration.duration,
                Exporter.date_to_string(case.compare_duration.start),
                Exporter.date_to_string(case.compare_duration.end),
                case.compare_duration.duration,
                case.error_type,
                case.error_message,
            ]

            data.append(case_data)

            if case.df_compare_gap is not None:
                detail_file_path = f"{self.output_path}/json/test_results/{name}.xlsx"

                os.makedirs(os.path.dirname(detail_file_path), exist_ok=True)
                case.df_compare_gap.to_excel(detail_file_path)

        os.makedirs(os.path.dirname(output_file), exist_ok=True)
        with open(output_file, "w", encoding="UTF-8") as f:
            writer = csv.writer(f, lineterminator="\n")
            writer.writerows(data)
            f.close()



"""Data connectors"""
from importlib import import_module
import inspect
import os

def get_connectors(spark_session):
    """Get all existing connectors"""

    connectors = {}

    files = [name for name in os.listdir(os.path.dirname(__file__)) if name.endswith(".py") and name.startswith("connector_")]
    for file in files:
        module_name = file[:-3]
        for name, obj in inspect.getmembers(import_module(f"connectors.{module_name}")):
            if inspect.isclass(obj) and name.startswith("Connector"):
                current_connector = obj()
                if spark_session is not None and current_connector.is_spark:
                    current_connector.spark = spark_session

                connectors[current_connector.name] = current_connector
    return connectors



# pylint: disable=R0903
"""Connector to read SQL file"""

from connectors.connector import Connector

class ConnectorSQLSpark(Connector):
    """Connector to execute SQL query over Spark"""
    def __init__(self):
        self.name = "SQL_SPARK"
        self.is_spark = True
        self.connection_definition = []
        self.configuration_definition = [
            { "name": "query" }
        ]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""

        df = self.spark.sql(configuration["query"])

        return df



# pylint: disable=R0903
"""Connector to read Databricks database"""

import pandas as pd
from sqlalchemy import create_engine
from connectors.connector import Connector

class ConnectorDatabricks(Connector):
    """Connector to read Databricks database"""

    def __init__(self):
        self.name = "DATABRICKS"
        self.connection_definition = [
            {
                "name": "token"
            },    
            {
                "name": "hostname"
            },
            {
                "name": "database"
            },
            {
                "name": "http_path"
            },
            {
                "name":"port",
                "default": 443,
                "type": "integer"
            }
        ]
        self.configuration_definition = [{ "name": "query" }, { "name": "connection" }]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""

        token = connection["token"]
        hostname = connection["hostname"]
        database = connection["database"]
        port = connection["port"]
        http_path = connection["http_path"]

        connection_string = f"databricks://token:{token}@{hostname}:{port}/{database}?http_path={http_path}"

        sql_connection = create_engine(connection_string, echo=False)

        df = pd.read_sql(configuration["query"], sql_connection)

        return df



# pylint: disable=R0903
"""Connector to read ODBC connection"""

import warnings
import pandas as pd
import pyodbc
from sqlalchemy import create_engine
from connectors.connector import Connector

class ConnectorODCB(Connector):
    """Connector to read ODBC connection"""

    def __init__(self):
        self.name = "ODBC"
        self.connection_definition = [
            {
                "name": "DSN"
            },
            {
                "name": "auto_commit",
                "type": "boolean",
                "default": True,
            },
            {
                "name": "use_credentials",
                "type": "boolean",
                "default": False,
            },
            {
                "name": "user",
                "default": None,
            },
            {
                "name": "password",
                "default": None,
            },
        ]
        self.configuration_definition = [{ "name": "query" }, { "name": "connection" }]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""

        if connection["use_credentials"]:
            odbc_connection = pyodbc.connect(f"DSN={connection['DSN']}", user=connection["user"], password=connection["password"], autocommit=connection["auto_commit"])
        else:
            odbc_connection = pyodbc.connect(f"DSN={connection['DSN']}", autocommit=connection["auto_commit"])

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
          
            df = pd.read_sql(configuration["query"], odbc_connection)

        return df



"""Module to manage the configuration of the framework"""
import copy
import pathlib
import yaml
from pyjeb import control_and_setup
from pyjeb.exception import InvalidParameterException
from case import Case, ConnectionDescription
from parameters import Parameters

class Configuration:
    """Manage configuration"""
    parameters = None
    connectors = None
    exporter = None
    connections = {}

    case_definition = [
        { "name": "disabled", "type": "boolean", "default": False },
        { "name": "options", "type": "dict", "default": {} },
        { "name": "options.ignore", "type": "list", "default": None },
        { "name": "options.sort", "type": "list", "default": None },
        { "name": "source", "type": "dict" },
        { "name": "source.type" },
        { "name": "source.connection", "default": None },
        { "name": "expected", "type": "dict" },
        { "name": "expected.connection", "default": None },
        { "name": "expected.connection", "default": None },
    ]

    def __init__(self, parameters:Parameters, connectors:dict, exporters:dict):
        self.parameters = parameters
        self.connectors = connectors

        self.init_connections()
        self.set_exporter(exporters)

    def get_module_type(self, obj:dict):
        """Get type of module from user configuration"""
        if "type" not in obj.keys():
            raise InvalidParameterException("Property 'type' is not defined")

        module_type = str(obj["type"]).upper()
        if module_type not in self.connectors.keys():
            raise InvalidParameterException(f"Property 'type' ({module_type}) is not a valid connector")

        return module_type

    def get_connection(self, name):
        """Get connection of module from user configuration"""

        if name is None:
            return None

        if name not in self.connections.keys():
            raise InvalidParameterException(f"Connection {name} do not exists")

        return self.connections[name]

    def init_connections(self):
        """Load connections and apply control and variable setup"""

        if self.parameters.path_connection is None:
            self.connections = {}
        else:
            with open(self.parameters.path_connection, encoding="UTF-8") as file:
                content = file.read()
                self.connections = yaml.safe_load(content)

        connections = copy.copy(self.connections)
        for connection_name in connections:
            connection = connections[connection_name]
            module_type = self.get_module_type(connection)

            connection_definition = self.connectors[module_type].connection_definition

            self.connections[module_type] = control_and_setup(connection, connection_definition, self.parameters.variables, None)

    def get_cases(self):
        """Get cases from user configuration and apply control and variable setup"""

        cases = {}
        files_list = list(pathlib.Path(self.parameters.path_cases).rglob(self.parameters.path_cases_filter))

        for file_path in files_list:
            with open(file_path, encoding="UTF-8") as file:
                configurations = yaml.load(file, Loader = yaml.loader.SafeLoader)
                for case_name in configurations.keys():
                    configuration = configurations[case_name]

                    case = control_and_setup(configuration, self.case_definition, self.parameters.variables, None)

                    source_module_type = self.get_module_type(configuration["source"])
                    source_connector = self.connectors[source_module_type]
                    source_configuration_definition = source_connector.configuration_definition

                    case["source"] = control_and_setup(case["source"], source_configuration_definition, self.parameters.variables, None)
                    source_definition = ConnectionDescription(source_connector, self.get_connection(case["source"]["connection"]))

                    expected_module_type = self.get_module_type(configuration["expected"])
                    expected_connector = self.connectors[expected_module_type]
                    expected_configuration_definition = expected_connector.configuration_definition

                    case["expected"] = control_and_setup(case["expected"], expected_configuration_definition, self.parameters.variables, None)
                    expected_definition = ConnectionDescription(expected_connector, self.get_connection(case["expected"]["connection"]))

                    if source_connector.is_spark != expected_connector.is_spark:
                        raise InvalidParameterException(f"'{case_name}': Source and expected must be both Spark or not Spark connectors")

                    cases[case_name] = Case(configuration, source_definition, expected_definition, case["options"], case["disabled"])

        return cases

    def set_exporter(self, exporters):
        """Set exporter from args"""

        if self.parameters.export not in exporters.keys():
            raise InvalidParameterException(f"Parameter 'export' ({self.parameters.export}) is not a valid exporter")

        self.exporter = exporters[self.parameters.export]
        self.exporter.output_path = self.parameters.path_output



# pylint: disable=R0903
"""Connector to read BigQuery database"""

import pandas as pd
import pandas_gbq
from sqlalchemy import create_engine
from connectors.connector import Connector

class ConnectorBigQuery(Connector):
    """Connector to read BigQuery database"""

    def __init__(self):
        self.name = "BIGQUERY"
        self.connection_definition = [
            {
                "name": "credentials",
                "default": None
            },
            {
                "name": "credentials_type",
                "validset": ["service_account", "current_user"],
                "default": "service_account"
            },
            {
                "name": "project_id",
                "default": None
            }
        ]
        self.configuration_definition = [
            { "name": "query" }
            , { "name": "connection" }
        ]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""
        credentials = connection["credentials"]
        credentials_type = connection["credentials_type"]

        if credentials_type == "service_account":
            connection_string = f"bigquery://?credentials_base64={credentials}"
            sql_connection = create_engine(connection_string, echo=False)
            df = pd.read_sql(configuration["query"], sql_connection)
        elif credentials_type == "current_user":
            df = pandas_gbq.read_gbq(configuration["query"], connection["project_id"], progress_bar_type = None)

        return df



# pylint: disable=R0903
"""Connector to return empty"""

import pandas as pd
from pyspark.sql.types import StructType
from connectors.connector import Connector

class ConnectorEmpty(Connector):
    """Connector to return empty"""
    def __init__(self):
        self.name = "EMPTY_SPARK"
        self.is_spark = True
        self.connection_definition = []
        self.configuration_definition = []

    def get_data(self, configuration: dict, connection: dict):
        """Return empty value"""
        empty_RDD = self.spark.sparkContext.emptyRDD()
        columns = StructType([])
        df = self.spark.createDataFrame(data = empty_RDD, schema = columns)

        return df



# pylint: disable=R0903
"""Connector to read PostgreSQL database"""

import pandas as pd
from sqlalchemy import create_engine
from connectors.connector import Connector

class ConnectorPostgreSQL(Connector):
    """Connector to read PostgreSQL database"""

    def __init__(self):
        self.name = "POSTGRESQL"
        self.connection_definition = [
            {
                "name": "mode",
                "default" : "password",
                "validset": ["password", "connection_string"]
            },    
            {
                "name": "hostname",
                "default" : None
            },
            {
                "name": "database",
                "default" : None
            },
            {
                "name": "username",
                "default" : None
            },
            {
                "name": "password",
                "default" : None
            },
            {
                "name":"port",
                "default": 5432,
                "type": "integer"
            },
             {
                "name":"ssl_context",
                "default": False,
                "type": "boolean"
            },           
            {
                "name":"connection_string",
                "default": None
            }
        ]
        self.configuration_definition = [{ "name": "query" }, { "name": "connection" }]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""

        connection_string = connection["connection_string"]
        if connection["mode"] == "password":
            port = connection["port"]
            hostname = connection["hostname"]
            username = connection["username"]
            password = connection["password"]
            database = connection["database"]
            connection_string = f"postgresql+pg8000://{username}:{password}@{hostname}:{port}/{database}"

        connect_args = {}
        if connection["ssl_context"] :
            connect_args = {'ssl_context': True}

        sql_connection = create_engine(connection_string, echo=False, connect_args=connect_args)

        df = pd.read_sql(configuration["query"], sql_connection)

        return df



# pylint: disable=R0903,W0613
"""Test result exporter"""
from datetime import datetime

class Exporter:
    """Test result exporter"""
    name = None
    output_path = None

    @staticmethod
    def date_to_string(data):
        """Convert datetime ton string"""
        if not isinstance(data, datetime):
            return None

        return data.strftime("%Y-%m-%dT%H:%M:%SZ")

    def export(self, cases:dict):
        """Export to destination"""
        return None



"""Export test case result to TRX format"""

import html
import os
import uuid
import xml.dom.minidom
import numpy as np
from exporters.exporter import Exporter
from case import StateStatistics 

class ExporterTRX(Exporter):
    """Export test case result to TRX format"""

    def __init__(self):
        self.name = "TRX"

    def get_failed_blocks(self, case_name, current_case, execution_id, output_folder):
        """Get XML code for failed cases"""
        error_message = html.escape(current_case.error_message, quote = False)

        output_message_xml = f"<Output><ErrorInfo><Message>{error_message}</Message></ErrorInfo></Output>"
        result_files_xml = ""
        
        if current_case.df_compare_gap is not None:
            detail_file_path = f"{output_folder}/test_results/In/{execution_id}/{case_name}.xlsx"
            result_files_xml = f"<ResultFiles><ResultFile path='{case_name}.xlsx'/></ResultFiles>"

            os.makedirs(os.path.dirname(detail_file_path), exist_ok=True)
            current_case.df_compare_gap.to_excel(detail_file_path)

        return output_message_xml, result_files_xml


    def export(self, cases:dict):
        """Export test case"""

        trx_id = str(uuid.uuid4())

        output_folder = f"{self.output_path}/trx"
        output_file = f"{output_folder}/test_results.xml"

        test_list_id = str(uuid.uuid4())

        execution_id_list = []
        test_id_list = []

        for i in list(range(0, len(cases))):
            execution_id_list.append(str(uuid.uuid4()))
            test_id_list.append(str(uuid.uuid4()))


        xml_unit_test_result = ""
        xml_test_definitions = ""
        xml_test_entry = ""

        state_statistics = StateStatistics()

        start_times = []
        end_times = []
        for i, case_name in enumerate(cases):
            current_case = cases[case_name]

            if current_case.global_duration.start is not None:
                start_times.append(current_case.global_duration.start)
                end_times.append(current_case.global_duration.end)

            execution_id = execution_id_list[i]
            test_id = test_id_list[i]

            state_statistics.add_state(current_case.state)

            output_message_xml = ""
            result_files_xml = ""

            if current_case.state != "passed" and current_case.error_message is not None:
                output_message_xml, result_files_xml = self.get_failed_blocks(case_name, current_case, execution_id_list[i], output_folder)

            outcome = current_case.state
            if outcome == "error":
                outcome = "failed"

            xml_unit_test_result += f"""<UnitTestResult
                executionId='{execution_id}'
                testId='{test_id}'
                testName='{case_name}'
                duration='{current_case.global_duration.duration}'
                startTime='{Exporter.date_to_string(current_case.global_duration.start)}'
                endTime='{Exporter.date_to_string(current_case.global_duration.end)}'
                outcome='{outcome}'
                testListId='{test_list_id}'>{output_message_xml}{result_files_xml}</UnitTestResult>"""

            xml_test_definitions += f"<UnitTest id='{test_id}' name='{case_name}'><Execution id='{execution_id}'/></UnitTest>"
            xml_test_entry += f"<TestEntry testId='{test_id}' executionId='{execution_id}' testListId='{test_list_id}'/>"

        global_start_date = Exporter.date_to_string(np.min(np.array(start_times)))
        global_end_date = Exporter.date_to_string(np.max(np.array(end_times)))

        xml_string = f"""<?xml version='1.0' encoding='UTF-8'?>
            <TestRun xmlns='http://microsoft.com/schemas/VisualStudio/TeamTest/2010' id='{trx_id}'>
                <Times creation='{global_start_date}' queueing='{global_start_date}' start='{global_start_date}' finish='{global_end_date}' />
                <TestSettings id='{trx_id}'/>
                <Results>{xml_unit_test_result}</Results>
                <TestDefinitions>{xml_test_definitions}</TestDefinitions>
                <TestEntries>{xml_test_entry}</TestEntries>
                <TestLists><TestList id='{test_list_id}' name='All Loaded Results'/></TestLists>
                <ResultSummary outcome='Complete'>
                    <Counters 
                        total='{state_statistics.total}'
                        executed='{state_statistics.executed}'
                        passed='{state_statistics.passed}'
                        failed='{state_statistics.failed}'
                        error='{state_statistics.error}'
                        notExecuted='{state_statistics.not_executed}' />
                    <Output StdOut='' />
                </ResultSummary>
            </TestRun>"""

        os.makedirs(os.path.dirname(output_file), exist_ok=True)
        with open(output_file, "w", encoding="UTF-8") as file:
            dom_string = xml.dom.minidom.parseString(xml_string).toprettyxml()
            dom_string = os.linesep.join([s for s in dom_string.splitlines() if s.strip()])
            file.write(dom_string)


"""Module to manage test case"""
from dataclasses import dataclass
from datetime import datetime
import numpy as np
import pandas as pd

@dataclass
class StateStatistics:
    """Statistics of test case executions"""
    not_executed = 0
    executed = 0
    passed = 0
    failed = 0
    error = 0
    total = 0

    def add_state(self, state):
        """add new state to statistics"""
        if state == "passed": self.passed += 1
        if state == "failed": self.failed += 1
        if state == "error": self.error += 1
        if state == "notExecuted": self.not_executed += 1

        if state != "notExecuted": self.executed += 1

        self.total += 1


@dataclass
class ConnectionDescription:
    """Tuple of connection and connector"""
    connector = None
    connection = None

    def __init__(self, connector, connection):
        self.connector = connector
        self.connection = connection

@dataclass
class Duration:
    """Structure of duration"""

    start = None
    end = None
    duration = None

    def calculate_duration(self):
        """Calculate the duration between start and end date"""
        if self.end is not None:
            duration = self.end - self.start
            self.duration = duration.seconds + (duration.microseconds / 1000000)


class CaseItem:
    """Structure of case item (source or expected)"""

    connector = None
    connection = None
    configuration = None

    duration = None
    df_data = None

    def __init__(self, configuration, connector, connection):
        self.duration = Duration()
        self.connector = connector
        self.connection = connection
        self.configuration = configuration

class Case:
    """Test case item"""

    options = None

    source = None
    expected = None

    global_duration = None
    compare_duration = None

    state = "notExecuted"
    error_type = None
    error_message = None

    df_compare_gap = None

    disabled = None

    def __init__(self, configuration, source, expected, options, disabled):
        self.source = CaseItem(configuration["source"], source.connector, source.connection)
        self.expected = CaseItem(configuration["expected"], expected.connector, expected.connection)
        self.options = options
        self.disabled = disabled
        self.global_duration = Duration()
        self.compare_duration = Duration()

    def load_data(self, obj_type:str):
        """ Load data from connector"""

        if obj_type == "source":
            obj = self.source
        else:
            obj = self.expected

        obj.duration.start = datetime.now()
        obj.df_data = obj.connector.get_data(obj.configuration, obj.connection)
        obj.duration.end = datetime.now()

    def load_data_error(self, obj_type:str, message:str):
        """Setup error message for data loading"""

        if obj_type == "source":
            obj = self.source
        else:
            obj = self.expected

        self.state = "error"
        self.error_type = "data"
        self.error_message = message
        obj.duration.end = datetime.now()

    def compare_dataframes(self):
        """Compare source and exptected dataframe"""

        self.compare_duration.start = datetime.now()

        count_source = len(self.source.df_data)
        count_expected = len(self.expected.df_data)

        if count_source != count_expected:
            self.error_type = "count"
            self.error_message = f"The count in source dataset ({count_source}) is differant than the count in the expected dataset ({count_expected})"

        if self.error_message is None and count_source != 0:
            if self.options is not None and self.options["ignore"] is not None:
                self.source.df_data = self.source.df_data.drop(columns = self.options["ignore"], axis = 1, errors = "ignore")
                self.expected.df_data = self.expected.df_data.drop(columns = self.options["ignore"], axis = 1, errors = "ignore")

            df_columns_source = pd.DataFrame({ "columns": self.source.df_data.columns }).sort_values(by = ["columns"]).reset_index(drop=True)
            df_columns_expected = pd.DataFrame({ "columns": self.expected.df_data.columns }).sort_values(by = ["columns"]).reset_index(drop=True)

            message = "The headers are differant between source dataset and expected dataset"
            if len(df_columns_source) != len(df_columns_expected):
                self.error_message = message
                self.error_type = "headers"
            else:
                df_compare = df_columns_source.compare(df_columns_expected, result_names = ("source", "expected"))
                if len(df_compare) != 0 :
                    self.error_message = message
                    self.error_type = "headers"
                    self.df_compare_gap = df_compare

        if self.error_message is None and count_source != 0:
            if self.options is not None and self.options["sort"] is not None:
                self.source.df_data = self.source.df_data.sort_values(by = self.options["sort"]).reset_index(drop=True)
                self.expected.df_data = self.expected.df_data.sort_values(by = self.options["sort"]).reset_index(drop=True)

            df_compare = self.source.df_data.compare(self.expected.df_data, result_names = ("source", "expected"))
            if len(df_compare) != 0 :
                self.error_message = "Some rows are not equals between source dataset and expected dataset"
                self.error_type = "data"
                self.df_compare_gap = df_compare

        self.compare_duration.end = datetime.now()

        if self.error_message is None:
            self.state = "passed"
        else:
            self.state = "failed"

    def compare_dataframes_with_spark(self, spark_session):
        """Compare source and expected dataframe using Spark"""

        self.compare_duration.start = datetime.now()

        count_source = self.source.df_data.count()
        count_expected = self.expected.df_data.count()

        if count_source != count_expected:
            self.error_type = "count"
            self.error_message = f"The count in source dataset ({count_source}) is different than the count in the expected dataset ({count_expected})"

        if self.error_message is None and count_source != 0:
            if self.options is not None and self.options["ignore"] is not None:
                ignore_columns = self.options["ignore"]
                self.source.df_data = self.source.df_data.drop(*ignore_columns)
                self.expected.df_data = self.expected.df_data.drop(*ignore_columns)

            source_columns = sorted(self.source.df_data.columns)
            expected_columns = sorted(self.expected.df_data.columns)

            if source_columns != expected_columns:
                self.error_message = "The headers are different between source dataset and expected dataset"
                self.error_type = "headers"

            if self.error_message is None and count_source != 0:
                #if self.options is not None and self.options["sort"] is not None:
                #    self.source.df_data = self.source.df_data.orderBy(*self.options["sort"]).drop("index")
                #    self.expected.df_data = self.expected.df_data.orderBy(*self.options["sort"]).drop("index")

                df_compare = self.source.df_data.exceptAll(self.expected.df_data)
                if df_compare.count() != 0:
                    self.error_message = "Some rows are not equal between source dataset and expected dataset"
                    self.error_type = "data"
                    self.df_compare_gap = df_compare.toPandas()

        self.compare_duration.end = datetime.now()

        if self.error_message is None:
            self.state = "passed"
        else:
            self.state = "failed"

    def compare_dataframes_error(self, message):
        """Setup error message for compare engine"""

        self.state = "Error"
        self.error_type = "compare"
        self.error_message = message
        self.compare_duration.end = datetime.now()

    def calculate_durations(self):
        """Calculate of durations"""

        self.source.duration.calculate_duration()
        self.expected.duration.calculate_duration()
        self.compare_duration.calculate_duration()

        ends = []
        if self.source.duration.end is not None: ends.append(self.source.duration.end)
        if self.expected.duration.end is not None: ends.append(self.expected.duration.end)
        if self.compare_duration.end is not None: ends.append(self.compare_duration.end)

        if len(ends) == 0:
            self.global_duration.duration = 0
        else:
            self.global_duration.start = self.source.duration.start
            self.global_duration.end = np.max(np.array(ends))
        self.global_duration.calculate_duration()



# pylint: disable=R0903
"""Connector to read Snowflake database"""

import pandas as pd
from sqlalchemy import create_engine
from connectors.connector import Connector

class ConnectorSnowflake(Connector):
    """Connector to read Snowflake database"""

    def __init__(self):
        self.name = "SNOWFLAKE"
        self.connection_definition = [
            {
                "name": "account_identifier"
            },
            {
                "name": "username"
            },
            {
                "name": "password"
            },
            {
                "name": "database",
                "default": None
            },
            {
                "name": "schema",
                "default": None
            },
            {
                "name": "warehouse",
                "default": None
            },
            {
                "name": "role",
                "default": None
            }
        ]
        self.configuration_definition = [{ "name": "query" }, { "name": "connection" }]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""


        account_identifier = connection["account_identifier"]
        username = connection["username"]
        password = connection["password"]
        connection_string = f"snowflake://{username}:{password}@{account_identifier}/"


        if connection["database"] is not None: connection_string += f"{connection['database']}/"
        if connection["schema"] is not None: connection_string += f"{connection['schema']}"

        connection_string += "?1=1"

        if connection["warehouse"] is not None: connection_string += f"&warehouse={connection['warehouse']}"
        if connection["role"] is not None: connection_string += f"&role={connection['role']}"

        sql_connection = create_engine(connection_string, echo=False)

        df = pd.read_sql(configuration["query"], sql_connection)

        return df



"""Result exporter"""
from importlib import import_module
import os
import inspect

def get_exporters():
    """Get all existings exporters"""
    connectors = {}

    files = [name for name in os.listdir(os.path.dirname(__file__)) if name.endswith(".py") and name.startswith("exporter_")]
    for file in files:
        module_mame = file[:-3]
        for name, obj in inspect.getmembers(import_module(f"exporters.{module_mame}")):
            if inspect.isclass(obj) and name.startswith("Exporter"):
                current_connector = obj()
                connectors[current_connector.name] = current_connector
    return connectors



# pylint: disable=R0903
"""Connector to read CSV file"""

from connectors.connector import Connector

class ConnectorCSVSpark(Connector):
    """Connector to read CSV file with Spark"""
    def __init__(self):
        self.name = "CSV_SPARK"
        self.is_spark = True
        self.connection_definition = []
        self.configuration_definition = [
            { "name": "path" },
            { "name": "delimiter", "default": "," },
            { "name": "header", "type": "boolean", "default": True }
        ]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""

        df = self.spark.read \
            .option("delimiter", configuration["delimiter"]) \
            .option("header", configuration["header"]) \
            .csv(configuration["path"])

        return df



"""Setup PyPi module"""
# pylint: disable=C0103

from setuptools import setup, find_packages
from ploosh.version import PLOOSH_VERSION

def setup_ploosh(name, install_requires):
    with open("../readme.md", encoding="UTF-8") as f:
        long_description = "".join(f.readlines())

    # replace relative link by absolute github link
    long_description = long_description.replace("(/", "(https://github.com/CSharplie/ploosh/blob/main/")

    install_requires = install_requires + [
            "colorama==0.4.6",
            "PyYAML==6.0.1",
            "Pyjeb==0.2.1",
            "numpy==1.26.3",
            "pandas==2.1.4",
            "openpyxl==3.1.2",
            "sqlalchemy==1.4.51",
            "pyspark==3.5.1",
        ]

    setup (
        name = name,
        version = PLOOSH_VERSION,
        description="A framework to automatize your tests for data projects",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/CSharplie/ploosh/",
        project_urls={
            "Bug Tracker": "https://github.com/CSharplie/ploosh/issues",
            "CI": "https://github.com/CSharplie/ploosh/actions",
            "Documentation": "https://github.com/CSharplie/ploosh",
            "Source Code": "https://github.com/CSharplie/ploosh",
        },
        download_url="https://pypi.org/project/ploosh/",
        platforms="Any",
        python_requires=">=3.6",
        license= "Apache License 2.0",
        entry_points = {
            "console_scripts": [
                "ploosh = ploosh.__main__:main"
            ]
        },
        install_requires=install_requires,
    )


"""Module for parsing input parameters"""

class Parameters:
    """Parse input parameters"""
    args = {}
    path_connection = None
    path_cases = None
    path_cases_filter = None
    path_output = None
    export = None
    failure_on_error = None
    variables = {}

    def __init__(self, argv:list):
        self.set_args(argv[1:])
        self.set_variables()
        self.path_connection = self.get_value("connections", None)
        self.path_cases = self.get_value("cases", "./cases")
        self.path_cases_filter = self.get_value("filter", "*.yml")
        self.path_output = self.get_value("output", "./output")
        self.export = self.get_value("export", "JSON").upper()
        self.failure_on_error = self.get_value("failure", True)
        self.spark_mode = self.get_value("spark", False)

    def set_args(self, args):
        """Set dictionary of args with cleaned name"""
        for i, name in enumerate(args):
            if not name.startswith("-"):
                continue

            value = False
            if i != len(args) - 1:
                value = args[i + 1]
                if value.startswith("-"):
                    value = True
                else:
                    value = value.replace("'", "").replace("\"", "")

            name = name.replace("-", "")
            self.args[name] = value

    def get_value(self, long_name:str, default):
        """Get value or default value from args"""
        if long_name in self.args:
            value = self.args[long_name]
            if str(value).upper() == "TRUE": return True
            if str(value).upper() == "FALSE": return False
            return value

        return default

    def set_variables(self):
        """Set variable liste from args"""
        for name, value in self.args.items():
            if not name.startswith("p_"):
                continue
            name = name.replace("p_", "")
            self.variables[name] = value



"""Export test case result to JSON format"""

import json
import os
from exporters.exporter import Exporter

class ExporterJSON(Exporter):
    """Export test case result to JSON format"""

    def __init__(self):
        self.name = "JSON"

    def export(self, cases:dict):
        """Export test case"""

        output_file = f"{self.output_path}/json/test_results.json"

        data = []
        for name in cases:
            case = cases[name]

            case_data = {
                "name": name,
                "state": case.state,
            }

            if case.source.duration.start is not None:
                case_data["source"] = {
                    "start": Exporter.date_to_string(case.source.duration.start),
                    "end":  Exporter.date_to_string(case.source.duration.end),
                    "duration":  case.source.duration.duration,
                }

            if case.expected.duration.start is not None:
                case_data["expected"] = {
                    "start": Exporter.date_to_string(case.expected.duration.start),
                    "end":  Exporter.date_to_string(case.expected.duration.end),
                    "duration":  case.expected.duration.duration,
                }

            if case.compare_duration.start is not None:
                case_data["compare"] = {
                    "start": Exporter.date_to_string(case.compare_duration.start),
                    "end": Exporter.date_to_string(case.compare_duration.end),
                    "duration": case.compare_duration.duration,
                }

            if case.state in ["error", "failed"]:
                case_data["error"] = {
                    "type": case.error_type,
                    "message": case.error_message
                }

            data.append(case_data)

            if case.df_compare_gap is not None:
                detail_file_path = f"{self.output_path}/json/test_results/{name}.xlsx"

                os.makedirs(os.path.dirname(detail_file_path), exist_ok=True)
                case.df_compare_gap.to_excel(detail_file_path)

        os.makedirs(os.path.dirname(output_file), exist_ok=True)
        with open(output_file, "w", encoding="UTF-8") as f:
            f.write(json.dumps(data, indent=2))


# pylint: disable=R0903
"""Connector to read Delta table"""

from connectors.connector import Connector

class ConnectorDeltaSpark(Connector):
    """Connector to read Delta table with Spark"""
    def __init__(self):
        self.name = "DELTA_SPARK"
        self.is_spark = True
        self.connection_definition = []
        self.configuration_definition = [
            { "name": "path" },
        ]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""

        df = self.spark.read.format("delta").load(configuration["path"])

        return df



from setup import setup_ploosh

install_requires = [
    "pyodbc==5.0.1",
    "pymysql==1.1.0",
    "pg8000==1.30.3",
    "snowflake-sqlalchemy==1.5.1",
    "databricks-sql-connector==2.9.3",
    "sqlalchemy-bigquery==1.9.0",
    "google-cloud-bigquery-storage==2.24.0",
    "pandas-gbq==0.23.0",
    "pydata-google-auth==1.8.2"
]

setup_ploosh("ploosh", install_requires)



# pylint: disable=R0903
"""Connector to read CSV file"""

import pandas as pd
from connectors.connector import Connector

class ConnectorCSV(Connector):
    """Connector to read CSV file"""
    def __init__(self):
        self.name = "CSV"
        self.connection_definition = []
        self.configuration_definition = [
            { "name": "path" },
            { "name": "delimiter", "default": "," }
        ]

    def get_data(self, configuration: dict, connection: dict):
        """Get data from source"""

        path = configuration["path"]
        df = pd.read_csv(path, delimiter=configuration["delimiter"])
        return df



# pylint: disable=R0903,W0613
"""Connector to access to remote data"""

class Connector:
    """Connector to access to remote data"""
    name = None
    connection_definition = None
    configuration_definition = None
    is_spark = False
    spark = None

    def get_data(self, configuration:dict, connection:dict):
        """Get data from connector"""
        return None
